// TODO: remove on implementation
#![allow(unused_imports)]
#![allow(unused_variables)]
#![allow(dead_code)]

use crate::model::*;

use std::collections::HashMap;
use std::fs;
use std::io;
use std::path::{Path, PathBuf};
use std::sync::Arc;

pub use self::search::Query;
mod collection_file;
mod image_convert;
mod search;

pub type TextureFileResult = Result<Arc<Vec<u8>>, TextureFileError>;
pub enum TextureFileError {
    NotFound,
    IoError(io::Error),
    ImageError(::image::ImageError),
}

pub struct DataStore {
    // private attributes
    // may change
    data_dir: PathBuf,
    texture: Vec<Texture>,
    preview_cache: HashMap<(TextureFormat, Sha256), Arc<Vec<u8>>>,
}

impl DataStore {
    pub fn new(path: &Path) -> io::Result<DataStore> {
        let base_path = path.to_path_buf();
        let mut collection_file_path = base_path.clone();
        collection_file_path.push("collection.json");

        Ok(DataStore {
            data_dir: base_path,
            texture: collection_file::load_collection_file(&collection_file_path)?,
            preview_cache: HashMap::new(),
        })
    }

    pub fn garbage_collect(&mut self) -> io::Result<()> {
        unimplemented!()
    }

    pub fn query(&mut self, query: &self::search::Query) -> Vec<Texture> {
        unimplemented!();
        // calls self::search::search(... )
    }

    /// returns true if successful
    pub fn delete(&mut self, tex: &Texture) -> bool {
        // Find the texture
        let pos = self.texture.iter().position(|e| e == tex);
        match pos {
            // Remove it
            Some(idx) => {
                let removed = self.texture.remove(idx);
                let key = (removed.format, removed.texture_hash);
                self.preview_cache.remove(&key);
                true
            }
            // Texture not found
            None => false,
        }
    }

    pub fn insert(&mut self, tex: Texture) -> bool {
        // Check for collisions
        let pos = self
            .texture
            .iter()
            .position(|e| e.id == tex.id || e.name == tex.name);
        match pos {
            // Texture with same name or id is already stored
            Some(_) => false,
            // Insert it
            None => {
                self.texture.push(tex);
                true
            }
        }
    }

    pub fn by_name<'a>(&'a self, name: &str) -> Option<&'a Texture> {
        let pos = self.texture.iter().position(|e| e.name == name);
        match pos {
            Some(idx) => Some(self.texture.get(idx).unwrap()),
            None => None,
        }
    }

    pub fn by_id<'a, 'b>(&'a self, id: &'b str) -> Option<&'a Texture> {
        let pos = self.texture.iter().position(|e| e.id == id);
        match pos {
            Some(idx) => Some(self.texture.get(idx).unwrap()),
            None => None,
        }
    }

    pub fn has_hash(&self, hash: &Sha256) -> bool {
        // TODO What is the purpose of this function?
        // Currently checks for textures in the list with the given hash.
        // Should it check for an existing image file instead?
        self.texture
            .iter()
            .position(|e| &e.texture_hash == hash)
            .is_some()
    }

    pub fn get_texture_file(&mut self, hash: &Sha256) -> TextureFileResult {
        use io::Read;

        let mut file_path = self.data_dir.clone();
        file_path.push("textures");
        file_path.push(hash.create_hex_string().to_lowercase());
        let mut file = match fs::File::open(file_path) {
            Ok(f) => f,
            Err(e) => {
                if e.kind() == io::ErrorKind::NotFound {
                    return Err(TextureFileError::NotFound);
                } else {
                    return Err(TextureFileError::IoError(e));
                }
            }
        };
        let mut buffer = Vec::new();
        match file.read_to_end(&mut buffer) {
            Ok(_) => (),
            Err(e) => return Err(TextureFileError::IoError(e)),
        };
        Ok(Arc::new(buffer))
    }

    pub fn get_texture_preview(&mut self, hash: &Sha256) -> TextureFileResult {
        unimplemented!();
    }
}
